/*
 * Copyright (c) 2020 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */

// CLASS HEADER
#include "color-visual.h"

// EXTERNAL INCLUDES
#include <dali/integration-api/debug.h>
#include <dali/devel-api/rendering/renderer-devel.h>

//INTERNAL INCLUDES
#include <dali-toolkit/public-api/visuals/color-visual-properties.h>
#include <dali-toolkit/public-api/visuals/visual-properties.h>
#include <dali-toolkit/devel-api/visuals/color-visual-properties-devel.h>
#include <dali-toolkit/devel-api/visuals/color-visual-actions-devel.h>
#include <dali-toolkit/internal/visuals/visual-factory-impl.h>
#include <dali-toolkit/internal/visuals/visual-factory-cache.h>
#include <dali-toolkit/internal/visuals/visual-string-constants.h>
#include <dali-toolkit/internal/visuals/visual-base-data-impl.h>

namespace Dali
{

namespace Toolkit
{

namespace Internal
{

namespace
{

const char* VERTEX_SHADER =
  "INPUT mediump vec2 aPosition;\n"

  "uniform highp mat4 uMvpMatrix;\n"
  "uniform highp vec3 uSize;\n"

  "//Visual size and offset\n"
  "uniform mediump vec2 offset;\n"
  "uniform highp vec2 size;\n"
  "uniform mediump vec4 offsetSizeMode;\n"
  "uniform mediump vec2 origin;\n"
  "uniform mediump vec2 anchorPoint;\n"
  "uniform mediump vec2 extraSize;\n"

  "vec4 ComputeVertexPosition()\n"
  "{\n"
  "  vec2 visualSize = mix(uSize.xy*size, size, offsetSizeMode.zw ) + extraSize;\n"
  "  vec2 visualOffset = mix( offset, offset/uSize.xy, offsetSizeMode.xy);\n"
  "  return vec4( (aPosition + anchorPoint)*visualSize + (visualOffset + origin)*uSize.xy, 0.0, 1.0 );\n"
  "}\n"

  "void main()\n"
  "{\n"
  "  gl_Position = uMvpMatrix * ComputeVertexPosition();\n"
  "}\n";


const char* FRAGMENT_SHADER =
  "uniform lowp vec4 uColor;\n"
  "uniform lowp vec3 mixColor;\n"

  "void main()\n"
  "{\n"
  "  OUT_COLOR = vec4(mixColor, 1.0) * uColor;\n"
  "}\n";

const char* VERTEX_SHADER_ROUNDED_CORNER =
  "INPUT mediump vec2 aPosition;\n"
  "OUTPUT mediump vec2 vPosition;\n"
  "OUTPUT mediump vec2 vRectSize;\n"
  "OUTPUT mediump float vCornerRadius;\n"

  "uniform highp mat4 uMvpMatrix;\n"
  "uniform highp vec3 uSize;\n"

  "//Visual size and offset\n"
  "uniform mediump vec2 offset;\n"
  "uniform highp vec2 size;\n"
  "uniform mediump vec2 extraSize;\n"
  "uniform mediump vec4 offsetSizeMode;\n"
  "uniform mediump vec2 origin;\n"
  "uniform mediump vec2 anchorPoint;\n"
  "uniform mediump float cornerRadius;\n"
  "uniform mediump float cornerRadiusPolicy;\n"

  "vec4 ComputeVertexPosition()\n"
  "{\n"
  "  vec2 visualSize = mix(uSize.xy*size, size, offsetSizeMode.zw ) + extraSize;\n"
  "  vec2 visualOffset = mix( offset, offset/uSize.xy, offsetSizeMode.xy);\n"
  "  mediump float minSize = min( visualSize.x, visualSize.y );\n"
  "  vCornerRadius = mix( cornerRadius * minSize, cornerRadius, cornerRadiusPolicy);\n"
  "  vCornerRadius = min( vCornerRadius, minSize * 0.5 );\n"
  "  vRectSize = visualSize / 2.0 - vCornerRadius;\n"
  "  vPosition = aPosition* visualSize;\n"
  "  return vec4( vPosition + anchorPoint*visualSize + (visualOffset + origin)*uSize.xy, 0.0, 1.0 );\n"
  "}\n"

  "void main()\n"
  "{\n"
  "  gl_Position = uMvpMatrix * ComputeVertexPosition();\n"
  "}\n";

//float distance = length( max( abs( position - center ), size ) - size ) - radius;
const char* FRAGMENT_SHADER_ROUNDED_CORNER =
  "INPUT mediump vec2 vPosition;\n"
  "INPUT mediump vec2 vRectSize;\n"
  "INPUT mediump float vCornerRadius;\n"

  "uniform lowp vec4 uColor;\n"
  "uniform lowp vec3 mixColor;\n"

  "void main()\n"
  "{\n"
  "  mediump float dist = length( max( abs( vPosition ), vRectSize ) - vRectSize ) - vCornerRadius;\n"
  "  OUT_COLOR = vec4(mixColor, 1.0) * uColor;\n"
  "  OUT_COLOR.a *= 1.0 - smoothstep( -1.0, 1.0, dist );\n"
  "}\n";

const char* VERTEX_SHADER_BLUR_EDGE =
  "INPUT mediump vec2 aPosition;\n"
  "OUTPUT mediump vec2 vPosition;\n"
  "OUTPUT mediump vec2 vRectSize;\n"

  "uniform highp   mat4 uMvpMatrix;\n"
  "uniform highp vec3 uSize;\n"

  "//Visual size and offset\n"
  "uniform mediump vec2 offset;\n"
  "uniform highp vec2 size;\n"
  "uniform mediump vec2 extraSize;\n"
  "uniform mediump vec4 offsetSizeMode;\n"
  "uniform mediump vec2 origin;\n"
  "uniform mediump vec2 anchorPoint;\n"
  "uniform mediump float blurRadius;\n"

  "vec4 ComputeVertexPosition()\n"
  "{\n"
  "  vec2 visualSize = mix(uSize.xy*size, size, offsetSizeMode.zw ) + extraSize + blurRadius * 2.0;\n"
  "  vec2 visualOffset = mix( offset, offset/uSize.xy, offsetSizeMode.xy);\n"
  "  vRectSize = visualSize / 2.0;\n"
  "  vPosition = aPosition* visualSize;\n"
  "  return vec4( vPosition + anchorPoint*visualSize + (visualOffset + origin)*uSize.xy, 0.0, 1.0 );\n"
  "}\n"

  "void main()\n"
  "{\n"
  "  gl_Position = uMvpMatrix * ComputeVertexPosition();\n"
  "}\n";

const char* FRAGMENT_SHADER_BLUR_EDGE =
  "INPUT mediump vec2 vPosition;\n"
  "INPUT mediump vec2 vRectSize;\n"

  "uniform lowp vec4 uColor;\n"
  "uniform lowp vec3 mixColor;\n"
  "uniform mediump float blurRadius;\n"

  "void main()\n"
  "{\n"
  "  mediump vec2 blur = 1.0 - smoothstep( vRectSize - blurRadius * 2.0, vRectSize, abs( vPosition ) );\n"
  "  OUT_COLOR = vec4(mixColor, 1.0) * uColor;\n"
  "  OUT_COLOR.a *= blur.x * blur.y;\n"
  "}\n";

}

ColorVisualPtr ColorVisual::New( VisualFactoryCache& factoryCache, const Property::Map& properties )
{
  ColorVisualPtr colorVisualPtr( new ColorVisual( factoryCache ) );
  colorVisualPtr->SetProperties( properties );
  return colorVisualPtr;
}

ColorVisual::ColorVisual(VisualFactoryCache& factoryCache)
: Visual::Base(factoryCache, Visual::FittingMode::FILL, Toolkit::Visual::COLOR),
  mBlurRadius(0.0f),
  mBlurRadiusIndex(Property::INVALID_INDEX),
  mRenderIfTransparent(false),
  mNeedBlurRadius(false)
{
}

ColorVisual::~ColorVisual()
{
}

void ColorVisual::DoSetProperties( const Property::Map& propertyMap )
{
  // By virtue of DoSetProperties being called last, this will override
  // anything set by Toolkit::Visual::Property::MIX_COLOR
  Property::Value* colorValue = propertyMap.Find( Toolkit::ColorVisual::Property::MIX_COLOR, MIX_COLOR );
  if( colorValue )
  {
    Vector4 color;
    if( colorValue->Get( color ) )
    {
      Property::Type type = colorValue->GetType();
      if( type == Property::VECTOR4 )
      {
        SetMixColor( color );
      }
      else if( type == Property::VECTOR3 )
      {
        Vector3 color3(color);
        SetMixColor( color3 );
      }
    }
    else
    {
      DALI_LOG_ERROR("ColorVisual: mixColor property has incorrect type\n");
    }
  }

  Property::Value* renderIfTransparentValue = propertyMap.Find( Toolkit::DevelColorVisual::Property::RENDER_IF_TRANSPARENT, RENDER_IF_TRANSPARENT_NAME );
  if( renderIfTransparentValue )
  {
    if( ! renderIfTransparentValue->Get( mRenderIfTransparent ) )
    {
      DALI_LOG_ERROR( "ColorVisual: renderIfTransparent property has incorrect type: %d\n", renderIfTransparentValue->GetType() );
    }
  }

  Property::Value* blurRadiusValue = propertyMap.Find( Toolkit::DevelColorVisual::Property::BLUR_RADIUS, BLUR_RADIUS_NAME );
  if( blurRadiusValue )
  {
    if( !blurRadiusValue->Get( mBlurRadius ) )
    {
      DALI_LOG_ERROR( "ColorVisual:DoSetProperties:: BLUR_RADIUS property has incorrect type: %d\n", blurRadiusValue->GetType() );
    }
  }
}

void ColorVisual::DoSetOnScene( Actor& actor )
{
  InitializeRenderer();

  // Only add the renderer if it's not fully transparent
  // We cannot avoid creating a renderer as it's used in the base class
  if( mRenderIfTransparent || mImpl->mMixColor.a > 0.0f )
  {
    actor.AddRenderer( mImpl->mRenderer );
  }

  // Color Visual generated and ready to display
  ResourceReady( Toolkit::Visual::ResourceStatus::READY );
}

void ColorVisual::DoSetOffScene(Actor& actor)
{
  if(mImpl->mRenderer && mBlurRadiusIndex != Property::INVALID_INDEX)
  {
    // Update values from Renderer
    mBlurRadius = mImpl->mRenderer.GetProperty<float>(mBlurRadiusIndex);
  }

  actor.RemoveRenderer(mImpl->mRenderer);
  mImpl->mRenderer.Reset();
  mBlurRadiusIndex = Property::INVALID_INDEX;
}

void ColorVisual::DoCreatePropertyMap( Property::Map& map ) const
{
  map.Clear();
  map.Insert( Toolkit::Visual::Property::TYPE, Toolkit::Visual::COLOR );
  map.Insert( Toolkit::ColorVisual::Property::MIX_COLOR, mImpl->mMixColor );
  map.Insert( Toolkit::DevelColorVisual::Property::RENDER_IF_TRANSPARENT, mRenderIfTransparent );

  if(mImpl->mRenderer && mBlurRadiusIndex != Property::INVALID_INDEX)
  {
    // Update values from Renderer
    float blurRadius = mImpl->mRenderer.GetProperty<float>(mBlurRadiusIndex);
    map.Insert(Toolkit::DevelColorVisual::Property::BLUR_RADIUS, blurRadius);
  }
  else
  {
    map.Insert(Toolkit::DevelColorVisual::Property::BLUR_RADIUS, mBlurRadius);
  }
}

void ColorVisual::DoCreateInstancePropertyMap( Property::Map& map ) const
{
  // Do nothing
}

void ColorVisual::OnSetTransform()
{
  if( mImpl->mRenderer )
  {
    mImpl->mTransform.RegisterUniforms( mImpl->mRenderer, Direction::LEFT_TO_RIGHT );
  }
}

void ColorVisual::OnDoAction( const Property::Index actionId, const Property::Value& attributes )
{
  // Check if action is valid for this visual type and perform action if possible
  switch( actionId )
  {
    case DevelColorVisual::Action::UPDATE_PROPERTY:
    {
      const Property::Map* map = attributes.GetMap();
      if( map )
      {
        DoSetProperties( *map );
      }
      break;
    }
  }
}

void ColorVisual::UpdateShader()
{
  if(mImpl->mRenderer)
  {
    Shader shader = GetShader();
    mImpl->mRenderer.SetShader(shader);
  }
}

void ColorVisual::InitializeRenderer()
{
  Geometry geometry = mFactoryCache.GetGeometry( VisualFactoryCache::QUAD_GEOMETRY );

  Shader shader = GetShader();

  mImpl->mRenderer = Renderer::New(geometry, shader);

  // ColorVisual has it's own index key for mix color - use this instead
  // of using the new base index to avoid changing existing applications
  // String keys will get to this property.
  mImpl->mMixColorIndex = mImpl->mRenderer.RegisterProperty(Toolkit::ColorVisual::Property::MIX_COLOR, MIX_COLOR, Vector3(mImpl->mMixColor));

  if(!EqualsZero(mBlurRadius))
  {
    mBlurRadiusIndex = mImpl->mRenderer.RegisterProperty(DevelColorVisual::Property::BLUR_RADIUS, BLUR_RADIUS_NAME, mBlurRadius);
    mImpl->mRenderer.SetProperty(Renderer::Property::BLEND_MODE, BlendMode::ON);
  }

  // Register transform properties
  mImpl->mTransform.RegisterUniforms(mImpl->mRenderer, Direction::LEFT_TO_RIGHT);
}

Shader ColorVisual::GetShader()
{
  Shader shader;
  if(!EqualsZero(mBlurRadius) || mNeedBlurRadius)
  {
    shader = mFactoryCache.GetShader( VisualFactoryCache::COLOR_SHADER_BLUR_EDGE );
    if( !shader )
    {
      shader = Shader::New( Dali::Shader::GetVertexShaderPrefix() + VERTEX_SHADER_BLUR_EDGE, Dali::Shader::GetFragmentShaderPrefix() + FRAGMENT_SHADER_BLUR_EDGE );
      mFactoryCache.SaveShader( VisualFactoryCache::COLOR_SHADER_BLUR_EDGE, shader );
    }
  }
  else if( !IsRoundedCornerRequired() )
  {
    shader = mFactoryCache.GetShader( VisualFactoryCache::COLOR_SHADER );
    if( !shader )
    {
      shader = Shader::New( Dali::Shader::GetVertexShaderPrefix() + VERTEX_SHADER, Dali::Shader::GetFragmentShaderPrefix() + FRAGMENT_SHADER );
      mFactoryCache.SaveShader( VisualFactoryCache::COLOR_SHADER, shader );
    }
  }
  else
  {
    shader = mFactoryCache.GetShader( VisualFactoryCache::COLOR_SHADER_ROUNDED_CORNER );
    if( !shader )
    {
      shader = Shader::New( Dali::Shader::GetVertexShaderPrefix() + VERTEX_SHADER_ROUNDED_CORNER, Dali::Shader::GetFragmentShaderPrefix() + FRAGMENT_SHADER_ROUNDED_CORNER );
      mFactoryCache.SaveShader( VisualFactoryCache::COLOR_SHADER_ROUNDED_CORNER, shader );
    }
  }

  return shader;
}

Dali::Property ColorVisual::OnGetPropertyObject(Dali::Property::Key key)
{
  if(!mImpl->mRenderer)
  {
    Handle handle;
    return Dali::Property(handle, Property::INVALID_INDEX);
  }

  if((key.type == Property::Key::INDEX && key.indexKey == DevelColorVisual::Property::BLUR_RADIUS) || (key.type == Property::Key::STRING && key.stringKey == BLUR_RADIUS_NAME))
  {
    mBlurRadiusIndex = mImpl->mRenderer.RegisterProperty(DevelColorVisual::Property::BLUR_RADIUS, BLUR_RADIUS_NAME, mBlurRadius);

    mImpl->mRenderer.SetProperty(Renderer::Property::BLEND_MODE, BlendMode::ON);

    mNeedBlurRadius = true;

    // Change shader
    UpdateShader();

    return Dali::Property(mImpl->mRenderer, mBlurRadiusIndex);
  }

  Handle handle;
  return Dali::Property(handle, Property::INVALID_INDEX);
}

} // namespace Internal

} // namespace Toolkit

} // namespace Dali
